# -*- coding: utf-8 -*-

import json
import logging
from collections import OrderedDict


import mamonsu.lib.platform as platform
if platform.PY3:
    import urllib.request as urllib2
else:
    import urllib2


class Request(object):

    def __init__(self, url, user, passwd):
        self._url, self._user, self._passwd = url, user, passwd
        self._id, self._auth_tocken = 0, None

    def _auth(self):
        if self._auth_tocken is None:
            self._auth_tocken = self.post(
                'user.login',
                {'user': self._user, 'password': self._passwd})
        return self._auth_tocken

    def _get_id(self):
        self._id += 1
        return self._id

    def _get_json(self, method, params):
        if method == 'user.login':
            data = OrderedDict([
                ('jsonrpc', '2.0'),
                ('method', method),
                ('params', params),
                ('id', self._get_id())
            ])
        else:
            data = OrderedDict([
                ('jsonrpc', '2.0'),
                ('method', method),
                ('params', params),
                ('auth', self._auth()),
                ('id', self._get_id())
            ])
        return json.dumps(data, sort_keys=False)

    def post(self, method, params):
        request = urllib2.Request(self._url)
        request.add_header('Content-Type', 'application/json; charset=utf-8')
        data = self._get_json(method, params)
        logging.debug('Post to zabbix: {0}'.format(data))
        if platform.PY3:
            data = bytearray(data, 'utf-8')
        response = urllib2.urlopen(request, data)
        logging.debug('Zabbix response code: {0}'.format(response.code))
        if response.code == 200:
            body = response.read()
            if platform.PY3:
                body = body.decode('utf-8')
            logging.debug('Zabbix response body: {0}'.format(body))
        else:
            raise Exception('Response code: {0}'.format(response.code))
        data = json.loads(body)
        if 'error' in data:
            raise Exception('Error response from zabbix api: {0}'.format(
                json.dumps(data['error'], indent=2)))
        return data['result']



# -*- coding: utf-8 -*-

import sys
import logging
import optparse
import os

from mamonsu import __version__
import mamonsu.lib.platform as platform
from mamonsu.lib.default_config import DefaultConfig
from mamonsu.plugins.pgsql.driver.checks import is_conn_to_db

from mamonsu.tools.report.pgsql import PostgresInfo
if platform.LINUX:
    from mamonsu.tools.report.os_linux import SystemInfo
else:
    if platform.WINDOWS:
        from mamonsu.tools.report.os_win import SystemInfo
if platform.LINUX:
    import pwd

reload(sys)
sys.setdefaultencoding('utf-8')


class Args(DefaultConfig):

    def __init__(self):

        parser = optparse.OptionParser(
            usage='%prog report',
            version='%prog report {0}'.format(__version__),
            description='First look report')
        group = optparse.OptionGroup(
            parser,
            'Start options')
        group.add_option(
            '--run-system',
            dest='run_system',
            default=True, help='Enable system collect (default: %default)')
        group.add_option(
            '--run-postgres',
            dest='run_postgres',
            default=True,
            help='Run postresql collect (default: %default)')
        group.add_option(
            '-l', '--log-level',
            dest='log_level',
            default='INFO', help='Log level (default: %default)')
        if platform.LINUX:
            group.add_option(
                '--disable-sudo',
                dest='disable_sudo', action='store_false',
                help='Disable sudo')
        parser.add_option_group(group)
        group = optparse.OptionGroup(
            parser,
            'Report options')
        group.add_option(
            '-w', '--report-path',
            dest='report_path',
            default=self.default_report_path(),
            help='Path to report (default: %default)')
        group.add_option(
            '-r', '--print-report',
            dest='print_report',
            default=True, help='Print summary info (default: %default)')
        parser.add_option_group(group)
        group = optparse.OptionGroup(
            parser,
            'Postgres connection options')
        group.add_option(
            '-d', '--dbname',
            dest='dbname',
            default=self.default_db(),
            help='database name to connect to (default: %default)')
        group.add_option(
            '--host',
            dest='hostname',
            default=self.default_host(),
            help='database server host or socket path (default: %default)')
        group.add_option(
            '--port',
            dest='port',
            default=self.default_port(),
            help='database server port (default: %default)')
        group.add_option(
            '-U', '--username',
            dest='username',
            default=self.default_user(),
            help='database user name (default: %default)')
        group.add_option(
            '-W', '--password',
            dest='password',
            default=self.default_user(),
            help='password (should happen automatically) ')
        parser.add_option_group(group)

        self.args, _ = parser.parse_args()

        # apply logging
        logging.basicConfig(
            filename=self.args.report_path,
            level=self.get_logger_level(self.args.log_level))
        # apply env
        os.environ['PGUSER'] = self.args.username
        os.environ['PGPASSWORD'] = self.args.password
        os.environ['PGHOST'] = self.args.hostname
        os.environ['PGDATABASE'] = self.args.dbname
        os.environ['PGAPPNAME'] = 'mamonsu first look'

    def try_configure_connect_to_pg(self):
        if not self._configure_auto_host():
            if self._try_run_as_postgres():
                if not self._configure_auto_host():
                    logging.error('Can\'t run as postgres')
                    self.run_postgres = False
            else:
                logging.error('Can\'t configure auto-host for postgresql')
                self.run_postgres = False

    def _configure_auto_host(self):
        if self.run_postgres and not self._auto_host_is_working():
            return False
        return True

    def _try_run_as_postgres(self):
        if platform.LINUX and os.getegid() == 0:
            try:
                uid = pwd.getpwnam('postgres').pw_uid
                os.seteuid(uid)
                return True
            except Exception as e:
                logging.error('Failed run as postgres: {0}'.format(e))
                pass
        return False

    def _auto_host_is_working(self):

        def test_db(self, host_pre):
            logging.debug('Test host: {0}'.format(host_pre))
            if is_conn_to_db(
                host=host_pre,
                db=self.args.dbname,
                port=self.args.port,
                user=self.args.username,
                    paswd=self.args.password):
                self.args.hostname = host_pre
                os.environ['PGHOST'] = self.args.hostname
                logging.debug('Connected via: {0}'.format(host_pre))
                return True
            return False

        host = self.args.hostname
        port = self.args.port
        if host == 'auto' and platform.UNIX:
            logging.debug('Host set to auto, test variables')
            if test_db(self, '/tmp/.s.PGSQL.{0}'.format(port)):
                return True
            if test_db(self, '/var/run/postgresql/.s.PGSQL.{0}'.format(port)):
                return True
            if test_db(self, '127.0.0.1'):
                return True
            # auto failed
            return False
        # not auto
        return True

    def __getattr__(self, name):
        try:
            return self.args.__dict__[name]
        except KeyError:
            return None


def run_report():

    args = Args()

    if args.run_system:
        sys_info = SystemInfo(args)
        sys_report = sys_info.collect()

    if args.run_postgres:
        args.try_configure_connect_to_pg()
        pg_info = PostgresInfo(args)
        pg_report = pg_info.collect()

    if args.print_report:
        if args.run_system:
            print(sys_report)
        else:
            print('Disabled system collect')
        if args.run_postgres:
            print(pg_report)
        else:
            print('Disabled postgres collect')



# -*- coding: utf-8 -*-
import sys
from mamonsu import __version__
from optparse import OptionParser, BadOptionError
import mamonsu.lib.platform as platform

usage_msg = """
Options:
    -a, --add-plugins  <directory>
    -c, --config       <file>
    -p, --pid          <pid-file>
    -d                 run in foreground

Export example config with default variables:
Command: export
Examples:
    {prog} export config <file>
Options:
    --add-plugins <directory>


Export zabbix template with additional plugins included in config file:
Command: export
Examples:
    {prog} export template <file>
Options:
    --add-plugins <directory>
    --template-name <template name>
    --application <application name in template>


Bootstrap DDL for monitoring:
Command: bootstrap
Examples:
    {prog} bootstrap <DBNAME>
Options:
    --port <PGPORT>
    -W <PGPASSWORD>
    -d, --dbname <DBNAME>
    -U, --username <USERNAME>


Information about working mamonsu:
Command: agent
Options:
    -c, --config <file>
Examples:
    {prog} agent version
    {prog} agent metric-list
    {prog} agent metric-get <metric key>


Zabbix API toolbox:
Command: zabbix
Options:
    --url=http://zabbix/web/face
    --user=WebUser
    --password=WebPassword
Examples:
    {prog} zabbix template list
    {prog} zabbix template show <template name>
    {prog} zabbix template id <template name>
    {prog} zabbix template delete <template id>
    {prog} zabbix template export <file>
    {prog} zabbix host list
    {prog} zabbix host show <host name>
    {prog} zabbix host id <host name>
    {prog} zabbix host delete <host id>
    {prog} zabbix host create <host name> <hostgroup id> <template id> <ip>
    {prog} zabbix host info templates <host id>
    {prog} zabbix host info hostgroups <host id>
    {prog} zabbix host info graphs <host id>
    {prog} zabbix host info items <host id>
    {prog} zabbix hostgroup list
    {prog} zabbix hostgroup show <hostgroup name>
    {prog} zabbix hostgroup id <hostgroup name>
    {prog} zabbix hostgroup delete <hostgroup id>
    {prog} zabbix hostgroup create <hostgroup name>
    {prog} zabbix item error <host name>
    {prog} zabbix item lastvalue <host name>
    {prog} zabbix item lastclock <host name>
"""


if platform.LINUX:
    usage_msg += """

Report about hardware and software:
Command: report
Options:
    --run-system
    --run-postgres
    --host <PGHOST>
    --port <PGPORT>
    -W <PGPASSWORD>
    -d, --dbname <DBNAME>
    -U, --username <USERNAME>
    -r, --print-report
    -w, --report-path <path to file>


AutoTune config and system:
Command: tune
Options:
    -l, --log-level INFO|DEBUG|WARN
    --dry-run
    --disable-sudo
    --dont-tune-pgbadger
    --dont-reload-postgresql
"""


def print_total_help():
    print(usage_msg.format(prog=sys.argv[0]))
    sys.exit(2)


class MissOptsParser(OptionParser):

    def print_help(self):
        print("""
Options:
    -c, --config <file>
    -p, --pid    <pid-file>

Export example config with default variables:
Command: export
Examples:
    {prog} export config <file>


Export zabbix template with additional plugins included in config file:
Command: export
Examples:
    {prog} export template <file>
Options:
    --config <file>
    -t <template name>
    -a <application name in template>
""")
        sys.exit(2)

    def _process_long_opt(self, rargs, values):
        try:
            OptionParser._process_long_opt(self, rargs, values)
        except BadOptionError as err:
            self.largs.append(err.opt_str)

    def _process_short_opts(self, rargs, values):
        try:
            OptionParser._process_short_opts(self, rargs, values)
        except BadOptionError as err:
            self.largs.append(err.opt_str)


def parse_args():
    parser = MissOptsParser(
        usage=usage_msg,
        version='%prog {0}'.format(__version__))
    parser.add_option(
        '-c', '--config', dest='config_file', default=None)
    # pid
    parser.add_option(
        '-p', '--pid', dest='pid', default=None)
    # daemonize
    parser.add_option(
        '-d', '--daemon', dest='daemon', action='store_true')
    # external plugins
    parser.add_option(
        '-a', '--add-plugins', dest='plugins_dirs', action='append')
    # template
    parser.add_option(
        '-t', '--template-name', dest='template',
        default='PostgresPro-{0}'.format(sys.platform.title()))
    parser.add_option(
        '--application', dest='application',
        default='App-PostgresPro-{0}'.format(sys.platform.title()))
    return parser.parse_args()



from mamonsu.plugins.system.plugin import Plugin
from .helpers import PerfData


class Memory(Plugin):

    Items = [
        # perf_item, zbx_key, desc, delta, unit, (color, site)
        (r'\Memory\Cache Bytes', '[cache]',
            'Memory cached', Plugin.UNITS.bytes, ('0000CC', 0)),

        (r'\Memory\Available Bytes', '[available]',
            'Memory available', Plugin.UNITS.bytes, ('00CC00', 0)),

        (r'\Memory\Free & Zero Page List Bytes', '[free]',
            'Memory free', Plugin.UNITS.bytes, ('CC0000', 0))
    ]

    def run(self, zbx):
        perf_services = []
        for _, item in enumerate(self.Items):
            perf_services.append(item[0])
        data = PerfData.get(perf_services, delay=1000)
        for idx, item in enumerate(self.Items):
            zbx.send('system.memory{0}'.format(item[1]), int(data[idx]))

    def items(self, template):
        result = ''
        for item in self.Items:
            result += template.item({
                'name': '{0}'.format(item[2]),
                'key': 'system.memory{0}'.format(item[1]),
                'units': item[3]
            })
        return result

    def graphs(self, template):
        items = []
        for item in self.Items:
            if item[4] is not None:
                items.append({
                    'key': 'system.memory{0}'.format(item[1]),
                    'color': item[4][0],
                    'yaxisside': item[4][1]
                })
        graph = {
            'name': 'Memory overview', 'items': items, 'type': 1}
        return template.graph(graph)



import mamonsu.lib.platform as platform
from distutils.version import LooseVersion
from ._connection import Connection, ConnectionInfo


class Pool(ConnectionInfo):

    ExcludeDBs = ['template0', 'template1', 'postgres']

    SQL = {
        # query type: ( 'if_not_installed', 'if_installed' )
        'replication_lag_master_query': (
            'select 1 as replication_lag_master_query',
            'select public.mamonsu_timestamp_master_update()'
        ),
        'replication_lag_slave_query': (
            'select extract(epoch from now()-pg_last_xact_replay_timestamp())',
            'select public.mamonsu_timestamp_get()'
        ),
        'count_xlog_files': (
            "select count(*) from pg_catalog.pg_ls_dir('pg_xlog')",
            'select public.mamonsu_count_xlog_files()'
        ),
        'count_autovacuum': (
            """select count(*) from pg_catalog.pg_stat_activity where
query like '%%autovacuum%%' and state <> 'idle'
and pid <> pg_catalog.pg_backend_pid()
        """,
            'select public.mamonsu_count_autovacuum()'
        ),
        'buffer_cache': (
            """select
sum(1) * 8 * 1024 as size,
sum(case when usagecount > 1 then 1 else 0 end) * 8 * 1024 as twice_used,
sum(case isdirty when true then 1 else 0 end) * 8 * 1024 as dirty
from public.pg_buffercache""",
            'select size, twice_used, dirty from public.mamonsu_buffer_cache()'
        ),
    }

    def __init__(self):
        super(Pool, self).__init__()
        self.all_connections = {}
        self._server_version = {}
        self._mamonsu_bootstrap = {}
        self._in_recovery = {}
        self._in_recovery_cache = 0

    def connection_string(self, db=None):
        self._init_conn_(db)
        return self.all_connections[db].conn_str()

    def query(self, query, db=None):
        if db is None:
            db = self.db
        self._init_conn_(db)
        return self.all_connections[db].query(query)

    def server_version(self, db=None):
        if db in self._server_version:
            return self._server_version[db]
        if platform.PY2:
            result = self.query('show server_version', db)[0][0]
        elif platform.PY3:
            result = bytes(
                self.query('show server_version', db)[0][0], 'utf-8')
        self._server_version[db] = "{0}".format(result.decode('ascii'))
        return self._server_version[db]

    def in_recovery(self, db=None):
        if (db in self._in_recovery) and (self._in_recovery_cache < 10):
            self._in_recovery_cache += 1
            return self._server_version[db]
        self._in_recovery_cache = 0
        self._in_recovery[db] = self.query(
            "select pg_catalog.pg_is_in_recovery()")[0][0]
        return self._in_recovery[db]

    def server_version_greater(self, version, db=None):
        return self.server_version(db) >= LooseVersion(version)

    def server_version_less(self, version, db=None):
        return self.server_version(db) <= LooseVersion(version)

    def mamonsu_bootstrap(self, db=None):
        if db in self._mamonsu_bootstrap:
            return self._mamonsu_bootstrap[db]
        sql = """select count(*) from pg_catalog.pg_class
            where relname = 'mamonsu_config'"""
        result = int(self.query(sql, db)[0][0])
        self._mamonsu_bootstrap[db] = (result == 1)
        if self._mamonsu_bootstrap[db]:
            self.log.info("Detect mamonsu bootstrap")
        else:
            self.log.info("Can't detect mamonsu bootstrap")
        return self._mamonsu_bootstrap[db]

    def extension_installed(self, ext, db=None):
        result = self.query('select count(*) from pg_catalog.pg_extension\
            where extname = \'{0}\''.format(ext), db)
        return (int(result[0][0])) == 1

    def databases(self):
        result, databases = self.query('select datname from \
            pg_catalog.pg_database'), []
        for row in result:
            if row[0] not in self.ExcludeDBs:
                databases.append(row[0])
        return databases

    def get_sql(self, typ, db=None):
        if typ not in self.SQL:
            raise LookupError("Unknown SQL type: '{0}'".format(typ))
        result = self.SQL[typ]
        if self.mamonsu_bootstrap(db):
            return result[1]
        else:
            return result[0]

    def run_sql_type(self, typ, db=None):
        return self.query(self.get_sql(typ, db), db)

    def _init_conn_(self, db):
        conn = self.all_connections.get(db)
        if conn is None:
            info = self._connection_info
            info['db'] = db
            self.all_connections[db] = Connection(info)



from mamonsu.lib.plugin import Plugin, PluginDisableException
from .pool import Pooler


class PgsqlPlugin(Plugin):

    is_child = False

    def __init__(self, config):
        super(PgsqlPlugin, self).__init__(config)
        self._enabled = config.fetch('postgres', 'enabled', bool)
        self._ext_installed, self._ext_check_count = False, None

    @classmethod
    def only_child_subclasses(self):
        # return all childs
        return self.__subclasses__()

    def extension_installed(self, ext, db=None, silent=False):

        def check(self, ext):
            self._ext_installed = Pooler.extension_installed(ext, db)
            self._ext_check_count = 0
            if not self._ext_installed and not silent:
                self.log.info("Extension '{0}' is not installed".format(ext))

        if self._ext_check_count is None:
            # first check
            check(self, ext)
        elif self._ext_check_count > 5:
            # try to RE-check
            check(self, ext)
        self._ext_check_count += 1

        return self._ext_installed

    def disable_and_exit_if_extension_is_not_installed(self, ext, db=None):
        if not self.extension_installed(ext, db=db, silent=True):
            self.disable()
            raise PluginDisableException("""
Disable plugin and exit, because '{0}' \
extension is not installed. Enable it in PostgreSQL instance: '{1}', \
if needed and restart.""".format(ext, Pooler.connection_string(db)))



# -*- coding: utf-8 -*-

import time

from mamonsu.lib.sender import *
from mamonsu.lib.senders import *
from mamonsu.tools.agent import *
from mamonsu.plugins import *


class Supervisor(object):

    Running = True

    def __init__(self, config):
        self.Plugins = []
        self.config = config
        self._sender = Sender()
        self._senders = []

    def start(self):
        self._load_plugins()
        self._update_senders()
        self._start_plugins()
        self._loop()

    def _load_plugins(self):
        for klass in Plugin.only_child_subclasses():
            plugin = klass(self.config)
            self.Plugins.append(plugin)

    def _update_senders(self):
        for plugin in self.Plugins:
            plugin.set_sender(self._sender)
            if plugin.is_sender():
                self._sender.add_sender(plugin)

    def _start_plugins(self):
        for plugin in self.Plugins:
            if plugin.is_enabled() and not plugin.is_alive():
                plugin.start()

    def _loop(self):
        plugin_errors, plugin_probes, last_error = 0, 0, ''
        while self.Running:
            for plugin in self.Plugins:
                if plugin.is_enabled() and not plugin.is_alive():
                    plugin.start()
                    last_error = plugin.last_error_text
                    plugin_errors += 1
            time.sleep(10)
            # error counts
            plugin_probes += 1
            if plugin_probes >= 60:
                if plugin_errors > 0:
                    self._sender.send(
                        'mamonsu.plugin.errors[]',
                        'Errors in the last 60 seconds: {0}.\
                        Last error: {1}'.format(
                            plugin_errors, last_error))
                else:
                    self._sender.send('mamonsu.plugin.errors[]', '')
                plugin_errors, plugin_probes = 0, 0



# -*- coding: utf-8 -*-

from mamonsu.plugins.pgsql.plugin import PgsqlPlugin as Plugin
from .pool import Pooler


class Xlog(Plugin):

    DEFAULT_CONFIG = {'lag_more_then_in_sec': str(60 * 5)}

    def run(self, zbx):
        if Pooler.in_recovery():
            # replication lag
            lag = Pooler.run_sql_type('replication_lag_slave_query')
            if lag[0][0] is not None:
                zbx.send('pgsql.replication_lag[sec]', float(lag[0][0]))
        else:
            Pooler.run_sql_type('replication_lag_master_query')
            # xlog location
            result = Pooler.query("""
                select pg_catalog.pg_xlog_location_diff
                    (pg_catalog.pg_current_xlog_location(),'0/00000000')""")
            zbx.send(
                'pgsql.wal.write[]', float(result[0][0]), self.DELTA_SPEED)
        # count of xlog files
        result = Pooler.run_sql_type('count_xlog_files')
        zbx.send('pgsql.wal.count[]', int(result[0][0]))

    def items(self, template):
        return template.item({
            'name': 'PostgreSQL: streaming replication lag',
            'key': 'pgsql.replication_lag[sec]'
        }) + template.item({
            'name': 'PostgreSQL: wal write speed',
            'key': 'pgsql.wal.write[]',
            'units': Plugin.UNITS.bytes
        }) + template.item({
            'name': 'PostgreSQL: count of xlog files',
            'key': 'pgsql.wal.count[]'
        })

    def graphs(self, template):
        result = template.graph({
            'name': 'PostgreSQL write-ahead log generation speed',
            'items': [
                {'color': 'CC0000',
                    'key': 'pgsql.wal.write[]'}]})
        result += template.graph({
            'name': 'PostgreSQL replication lag in second',
            'items': [
                {'color': 'CC0000',
                    'key': 'pgsql.replication_lag[sec]'}]})
        result += template.graph({
            'name': 'PostgreSQL count of xlog files',
            'items': [
                {'color': 'CC0000',
                    'key': 'pgsql.wal.count[]'}]})
        return result

    def triggers(self, template):
        return template.trigger({
            'name': 'PostgreSQL streaming lag to high '
            'on {HOSTNAME} (value={ITEM.LASTVALUE})',
            'expression': '{#TEMPLATE:pgsql.replication_lag[sec].last'
            '()}&gt;' + self.plugin_config('lag_more_then_in_sec')
        })



# -*- coding: utf-8 -*-

from mamonsu.plugins.pgsql.plugin import PgsqlPlugin as Plugin
from .pool import Pooler


class PgBufferCache(Plugin):

    Items = [
        # key, name, color
        ('size', 'PostgreSQL: shared buffer size', '0000CC'),
        ('twice_used', 'PostgreSQL: shared buffer twice used size', 'CC0000'),
        ('dirty', 'PostgreSQL: shared buffer dirty size', '00CC00')
    ]

    def run(self, zbx):
        if not self.extension_installed('pg_buffercache'):
            return
        result = Pooler.run_sql_type('buffer_cache')[0]
        for i, value in enumerate(result):
            zbx.send('pgsql.buffers[{0}]'.format(self.Items[i][0]), value)

    def items(self, template):
        result = ''
        for item in self.Items:
            result += template.item({
                'key': 'pgsql.buffers[{0}]'.format(item[0]),
                'name': item[1],
                'units': Plugin.UNITS.bytes
            })
        return result

    def graphs(self, template):
        items = []
        for item in self.Items:
            items.append({
                'key': 'pgsql.buffers[{0}]'.format(item[0]),
                'color': item[2]})
        return template.graph({
            'name': 'PostgreSQL: shared buffer',
            'items': items})



import os
from mamonsu.plugins.system.plugin import SystemPlugin as Plugin


class DiskSizes(Plugin):

    DEFAULT_CONFIG = {
        'vfs_percent_free': str(10),
        'vfs_inode_percent_free': str(10)}

    ExcludeFsTypes = [
        'none', 'unknown', 'rootfs', 'iso9660',
        'squashfs', 'udf', 'romfs', 'ramfs',
        'debugfs', 'cgroup', 'cgroup_root',
        'pstore', 'devtmpfs', 'autofs',
        'cgroup', 'configfs', 'devpts',
        'efivarfs', 'fusectl', 'fuse.gvfsd-fuse',
        'hugetlbfs', 'mqueue',
        'nfsd', 'proc', 'pstore',
        'rpc_pipefs', 'securityfs', 'sysfs',
        'nsfs', 'tmpfs', 'tracefs']

    def run(self, zbx):
        with open('/proc/self/mountinfo', 'r') as f:

            points = []

            for line in f:
                data = line.split()
                if len(data) != 11:
                    continue
                point, fstype = data[4], data[8]
                if fstype in self.ExcludeFsTypes:
                    continue
                try:
                    vfs = os.statvfs(point)
                except Exception as e:
                    self.log.error(
                        'Get statvfs for \'{0}\' error: {1}'.format(point, e))
                    continue
                if vfs.f_blocks == 0 or vfs.f_files == 0:
                    continue
                points.append({'{#MOUNTPOINT}': point})
                zbx.send(
                    'system.vfs.used[{0}]'.format(point),
                    int((vfs.f_blocks - vfs.f_bfree) * vfs.f_bsize))
                zbx.send(
                    'system.vfs.free[{0}]'.format(point),
                    int(vfs.f_bfree * vfs.f_bsize))
                zbx.send(
                    'system.vfs.percent_free[{0}]'.format(point),
                    100 - (float(vfs.f_blocks - vfs.f_bfree) * 100 / vfs.f_blocks))
                zbx.send(
                    'system.vfs.percent_inode_free[{0}]'.format(point),
                    100 - (float(vfs.f_files - vfs.f_ffree) * 100 / vfs.f_files))

            zbx.send('system.vfs.discovery[]', zbx.json({'data': points}))

    def discovery_rules(self, template):

        rule = {
            'name': 'VFS discovery',
            'key': 'system.vfs.discovery[]',
            'filter': '{#MOUNTPOINT}:.*'
        }

        items = [
            {
                'key': 'system.vfs.used[{#MOUNTPOINT}]',
                'name': 'Mount point {#MOUNTPOINT}: used',
                'value_type': Plugin.VALUE_TYPE.numeric_unsigned,
                'units': Plugin.UNITS.bytes},
            {
                'key': 'system.vfs.free[{#MOUNTPOINT}]',
                'name': 'Mount point {#MOUNTPOINT}: free',
                'value_type': Plugin.VALUE_TYPE.numeric_unsigned,
                'units': Plugin.UNITS.bytes},
            {
                'key': 'system.vfs.percent_free[{#MOUNTPOINT}]',
                'name': 'Mount point {#MOUNTPOINT}: free in percents',
                'units': Plugin.UNITS.percent},
            {
                'key': 'system.vfs.percent_inode_free[{#MOUNTPOINT}]',
                'name': 'Mount point {#MOUNTPOINT}: free inodes in percent',
                'units': Plugin.UNITS.percent}]

        graphs = [{
            'name': 'Mount point overview: {#MOUNTPOINT}',
            'type': self.GRAPH_TYPE.stacked,
            'items': [{
                    'color': 'CC0000',
                    'key': 'system.vfs.used[{#MOUNTPOINT}]'},
                {
                    'color': '0000CC',
                    'key': 'system.vfs.free[{#MOUNTPOINT}]'}]
        }]

        triggers = [{
            'name': 'Free disk space less then 10% on mountpoint '
            '{#MOUNTPOINT} (hostname={HOSTNAME} value={ITEM.LASTVALUE})',
            'expression': '{#TEMPLATE:system.vfs.'
            'percent_free[{#MOUNTPOINT}].last'
            '()}&lt;' + self.plugin_config('vfs_percent_free')},
            {
            'name': 'Free inode space less then 10% on mountpoint '
            '{#MOUNTPOINT} (hostname={HOSTNAME} value={ITEM.LASTVALUE})',
            'expression': '{#TEMPLATE:system.vfs.'
            'percent_inode_free[{#MOUNTPOINT}].last'
            '()}&lt;' + self.plugin_config('vfs_inode_percent_free')
        }]

        return template.discovery_rule(
            rule=rule, items=items, graphs=graphs, triggers=triggers)



# -*- coding: utf-8 -*-

from mamonsu.plugins.pgsql.plugin import PgsqlPlugin as Plugin
from .pool import Pooler


class PgWaitSampling(Plugin):

    AllLockItems = [
        # (sql_key, zbx_key, name, color)
        ('lwlock', 'all_lock[lwlock]',
            'Lightweight locks', '0000CC'),
        ('hwlock', 'all_lock[hwlock]',
            'Heavyweight locks', '00CC00'),
        ('buffer', 'all_lock[buffer]',
            'Buffer locks', 'CC0000')
    ]

    AllLockQuery = """
select
    CASE
        WHEN event_type = 'LWLockNamed' THEN 'lwlock'
        WHEN event_type = 'LWLockTranche' THEN 'lwlock'
        WHEN event_type = 'Lock' THEN 'hwlock'
        ELSE 'buffer'
    END,
    sum(count) as count
from pg_wait_sampling_profile
    where event_type is not null
group by 1
order by count desc"""

    HWLockItems = [
        # (sql_key, zbx_key, name, color)
        ('relation', 'hwlock[relation]',
            'lock on a relation', 'CC0000'),
        ('extend', 'hwlock[extend]',
            'extend a relation', '00CC00'),
        ('page', 'hwlock[page]',
            'lock on page', '0000CC'),
        ('tuple', 'hwlock[tuple]',
            'lock on a tuple', 'CC00CC'),
        ('transactionid', 'hwlock[transactionid]',
            'transaction to finish', '000000'),
        ('virtualxid', 'hwlock[virtualxid]',
            'virtual xid lock', 'CCCC00'),
        ('speculative token', 'hwlock[speculative_token]',
            'speculative insertion lock', '777777'),
        ('object', 'hwlock[object]',
            'lock on database object',
            '770000'),
        ('userlock', 'hwlock[userlock]',
            'userlock', '000077'),
        ('advisory', 'hwlock[advisory]',
            'advisory user lock', '007700')
    ]

    HWLockQuery = """
select
    event,
    sum(count) as count
from pg_wait_sampling_profile
    where event_type = 'Lock'
group by 1
order by count desc"""

    LWLockItems = [
        # (sql_key, zbx_key, name, color)
        ('xid', 'lwlock[xid]', 'XID access', 'BBBB00'),
        ('wal', 'lwlock[wal]', 'WAL access', 'CC0000'),
        ('clog', 'lwlock[clog]', 'CLOG access', '00CC00'),
        ('replication', 'lwlock[replication]', 'Replication Locks', 'FFFFCC'),
        ('buffer', 'lwlock[buffer]', 'Buffer operations', '0000CC'),
        ('other', 'lwlock[other]', 'Other operations', '007700')]

    LWLockQuery = """
select
    CASE
        WHEN event = 'ProcArrayLock' THEN 'xid'
        WHEN event = 'WALBufMappingLock' THEN 'wal'
        WHEN event = 'WALWriteLock' THEN 'wal'
        WHEN event = 'ControlFileLock' THEN 'wal'
        WHEN event = 'wal_insert' THEN 'wal'
        WHEN event = 'CLogControlLock' THEN 'clog'
        WHEN event = 'clog' THEN 'clog'
        WHEN event = 'SyncRepLock' THEN 'replication'
        WHEN event = 'ReplicationSlotAllocationLock' THEN 'replication'
        WHEN event = 'ReplicationSlotControlLock' THEN 'replication'
        WHEN event = 'ReplicationOriginLock' THEN 'replication'
        WHEN event = 'replication_origin' THEN 'replication'
        WHEN event = 'replication_slot_io' THEN 'replication'
        WHEN event = 'buffer_content' THEN 'buffer'
        WHEN event = 'buffer_io' THEN 'buffer'
        WHEN event = 'buffer_mapping' THEN 'buffer'
        ELSE 'other'
    END,
    sum(count) as count
from pg_wait_sampling_profile
    where event_type = 'LWLockTranche' or event_type = 'LWLockNamed'
group by 1
order by count desc"""

    def run(self, zbx):

        def find_and_send(where, what, zbx):
            for item in what:
                item_found = False
                for result in where:
                    if item[0] == result[0]:
                        zbx.send(
                            'pgsql.{0}'.format(item[1]),
                            float(result[1]), Plugin.DELTA.speed_per_second)
                        item_found = True
                        break
                if not item_found:
                    zbx.send(
                        'pgsql.{0}'.format(item[1]),
                        float(0), Plugin.DELTA.speed_per_second)

        self.disable_and_exit_if_extension_is_not_installed('pg_wait_sampling')

        find_and_send(Pooler.query(self.AllLockQuery), self.AllLockItems, zbx)
        find_and_send(Pooler.query(self.HWLockQuery), self.HWLockItems, zbx)
        find_and_send(Pooler.query(self.LWLockQuery), self.LWLockItems, zbx)

    def items(self, template):
        result = ''
        for item in (self.AllLockItems + self.LWLockItems + self.HWLockItems):
            result += template.item({
                'key': 'pgsql.{0}'.format(item[1]),
                'name': 'PostgreSQL waits: {0}'.format(item[2]),
                'value_type': self.VALUE_TYPE.numeric_float})
        return result

    def graphs(self, template):
        result = ''
        for graph_name, graph_items in [
                ('PostgreSQL waits: Locks by type', self.AllLockItems),
                ('PostgreSQL waits: Heavyweight locks', self.HWLockItems),
                ('PostgreSQL waits: Lightweight locks', self.LWLockItems)]:
            items = []
            for item in graph_items:
                items.append({
                    'key': 'pgsql.{0}'.format(item[1]),
                    'color': item[3]})
            result += template.graph({
                'name': graph_name, 'type': 1, 'items': items})
        return result



# -*- coding: utf-8 -*-
import json
import time

from mamonsu.lib.plugin import Plugin


class Sender():

    def __init__(self):
        self._senders = []
        self._last_values = {}

    def _hash(self, key, host=None):
        return '{0}_+_{1}'.format(host, key)

    def _key_from_hash(self, hash, host=None):
        result = hash.split('{0}_+_'.format(host))
        if len(result) == 1:
            return hash
        else:
            return result[1]

    def add_sender(self, sender):
        self._senders.append(sender)

    # resend all values to senders
    def send(self, key, value, delta=None, host=None, clock=None):

        if clock is None:
            clock = int(time.time())

        hash_key = self._hash(key, host)
        if delta is not None:
            if isinstance(value, float) or isinstance(value, int):
                if hash_key in self._last_values:
                    last_value, last_time = self._last_values[hash_key]
                    self._last_values[hash_key] = (value, clock)
                    if delta == Plugin.DELTA.speed_per_second:
                        value = float(value - last_value) / (clock - last_time)
                    if delta == Plugin.DELTA.simple_change:
                        value = float(value - last_value)
                else:
                    self._last_values[hash_key] = (value, clock)
                    return
        else:
            self._last_values[hash_key] = (value, clock)

        for sender in self._senders:
            if sender.is_enabled():
                sender.send(key, value, host, clock)

    # get last value: (value, clock)
    def get_metric(self, key, host=None):
        hash_key = self._hash(key, host)
        if hash_key in self._last_values:
            return self._last_values[hash_key]
        else:
            return (None, None)

    # list of metrics: [(key, (value, clock))]
    def list_metrics(self, host=None):
        hash_list, result = self._last_values.keys(), []
        for h in hash_list:
            result.append(
                (self._key_from_hash(h, host), self._last_values[h])
            )
        return result

    def json(self, val):
        return json.dumps(val)



# -*- coding: utf-8 -*-
import sys
import logging

import mamonsu.lib.platform as platform
from mamonsu.plugins.pgsql.pool import Pooler
from mamonsu.tools.sysinfo.linux import SysInfoLinux as SysInfo


class AutoTunePgsl(object):

    def __init__(self, args):

        if not self._is_connection_work():
            logging.error('Can\'t connect to PostgreSQL')
            sys.exit(5)

        self.args = args
        self.sys_info = SysInfo()

        self._memory()
        self._auto_vacuum()
        self._bgwriter()
        self._checkpointer()
        self._configure_pgbadger()
        self._configure_extensions()
        self._configure_virt_guest()

        self._reload_config()

    def _configure_extensions(self):

        extensions = self._run_query(
            "select name from pg_catalog.pg_available_extensions")
        if extensions is None:
            return
        extensions = [row[0] for row in extensions]

        needed_libraries = []

        if 'pg_stat_statements' in extensions:
            needed_libraries.append('pg_stat_statements')
        elif 'pg_buffercache' in extensions:
            needed_libraries.append('pg_buffercache')
        else:
            logging.warning("Please install 'contrib' modules: "
                            "need for 'pg_stat_statements'")

        if 'pg_wait_sampling' in extensions:
            needed_libraries.append('pg_wait_sampling')

        if len(needed_libraries) == 0:
            return

        libraries = self._run_query('show shared_preload_libraries;')
        if libraries is None:
            return
        elif not len(libraries) == 1:
            return
        elif not len(libraries[0]) == 1:
            return

        libraries = libraries[0][0]
        if len(libraries) == 0:
            libraries = needed_libraries
        else:
            libraries = libraries.split(',')
            libraries = [ext.strip() for ext in libraries]
            if needed_libraries not in libraries:
                for ext in needed_libraries:
                    if ext not in libraries:
                        libraries.append(ext)
        libraries = ','.join(libraries)
        self._run_query(
            "alter system set shared_preload_libraries to {0};".format(
                libraries))

    def _memory(self):
        if platform.WINDOWS:
            logging.info('No memory tune for windows')
            return

        sysmemory = self.sys_info.meminfo['_TOTAL']
        if sysmemory == 0:
            return

        self._run_query(
            "alter system set shared_buffers to '{0}';".format(
                self._humansize_and_round_bytes(sysmemory / 4)))
        self._run_query(
            "alter system set effective_cache_size to '{0}';".format(
                self._humansize_and_round_bytes(3 * sysmemory / 4)))
        self._run_query(
            "alter system set work_mem to '{0}';".format(
                self._humansize_and_round_bytes(sysmemory / 100)))
        self._run_query(
            "alter system set maintenance_work_mem to '{0}';".format(
                self._humansize_and_round_bytes(sysmemory / 10)))

    def _auto_vacuum(self):
        self._run_query(
            "alter system set autovacuum_max_workers to 20;")
        self._run_query(
            "alter system set autovacuum_analyze_scale_factor to 0.01;")
        self._run_query(
            "alter system set autovacuum_vacuum_scale_factor to 0.02;")
        self._run_query(
            "alter system set vacuum_cost_delay to 1;")

    def _bgwriter(self):
        self._run_query(
            "alter system set bgwriter_delay to 10;")
        self._run_query(
            "alter system set bgwriter_lru_maxpages to 800;")

    def _checkpointer(self):

        self._run_query(
            "alter system set checkpoint_completion_target to 0.75")

        sysmemory = self.sys_info.meminfo['_TOTAL']
        if sysmemory < 4 * 1024 * 1024 * 1024:
            return

        wal_size = min(sysmemory / 4, 8.0 * 1024 * 1024 * 1024)
        if Pooler.server_version_greater('9.5'):
            self._run_query(
                "alter system set max_wal_size to '{0}';".format(
                    self._humansize_and_round_bytes(wal_size)))

    def _configure_pgbadger(self):
        if self.args.pgbadger is not None:
            return
        self._run_query(
            "alter system set logging_collector to on;")
        self._run_query(
            "alter system set log_filename to 'postgresql-%%a.log';")
        self._run_query(
            "alter system set log_checkpoints to on;")
        self._run_query(
            "alter system set log_connections to on;")
        self._run_query(
            "alter system set log_disconnections to on;")
        self._run_query(
            "alter system set log_lock_waits to on;")
        self._run_query(
            "alter system set log_temp_files to 0;")
        self._run_query(
            "alter system set log_autovacuum_min_duration to 0;")
        self._run_query(
            "alter system set track_io_timing to on;")
        self._run_query(
            "alter system set log_line_prefix to "
            "'%%t [%%p]: [%%l-1] db=%%d,user=%%u,app=%%a,client=%%h ';")

    def _configure_virt_guest(self):
        if not self.sys_info.is_virt_guest():
            return
        self._run_query(
            "alter system set synchronous_commit to off;")

    def _reload_config(self):
        if self.args.reload_config is not None:
            return
        self._run_query('select pg_catalog.pg_reload_conf();')

    def _is_connection_work(self):
        try:
            Pooler.query('select 1')
            return True
        except Exception as e:
            logging.error('Test query error: {0}'.format(e))
            return False

    def _run_query(self, query='', exit_on_fail=True):
        if self.args.dry_run:
            logging.info('dry run (query):\t{0}'.format(
                query.replace('%%', '%')))
            return None
        try:
            return Pooler.query(query)
        except Exception as e:
            logging.error('Query {0} error: {1}'.format(query, e))
            if exit_on_fail:
                sys.exit(6)

    def _humansize_and_round_bytes(self, nbytes):
        suffixes = ['', 'kB', 'MB', 'GB', 'TB']
        if nbytes < 1024:
            return str(nbytes)
        i = 0
        while nbytes >= 1024 and i < len(suffixes) - 1:
            nbytes /= 1024.
            i += 1
        f = ('%.2f' % nbytes).rstrip('0').rstrip('.')
        return '%s%s' % (int(round(float(f))), suffixes[i])



__author__ = 'Dmitry Vasilyev'
__author_email__ = 'info@postgrespro.ru'
__description__ = 'Monitoring agent for PostgreSQL'
__version__ = '2.0.8'
__licence__ = 'BSD'

__url__ = 'https://github.com/postgrespro/mamonsu'

__classifiers__ = [
    'Environment :: Console',
    'Intended Audience :: System Administrators',
    'Operating System :: POSIX :: Linux',
    'Operating System :: Microsoft :: Windows',
    'License :: OSI Approved :: BSD License',
    'Programming Language :: Python :: 2',
    'Programming Language :: Python :: 3',
    'Topic :: System :: Monitoring'
]
__keywords__ = [
    'monitoring',
    'zabbix',
    'postgres'
]
